#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


PROJECT_ENV_NAME = "NEPTUNE_PROJECT"

API_TOKEN_ENV_NAME = "NEPTUNE_API_TOKEN"

CUSTOM_RUN_ID_ENV_NAME = "NEPTUNE_CUSTOM_RUN_ID"

MONITORING_NAMESPACE = "NEPTUNE_MONITORING_NAMESPACE"

NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE = "NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE"

NEPTUNE_NOTEBOOK_ID = "NEPTUNE_NOTEBOOK_ID"

NEPTUNE_NOTEBOOK_PATH = "NEPTUNE_NOTEBOOK_PATH"

NEPTUNE_RETRIES_TIMEOUT_ENV = "NEPTUNE_RETRIES_TIMEOUT"

NEPTUNE_SYNC_BATCH_TIMEOUT_ENV = "NEPTUNE_SYNC_BATCH_TIMEOUT"

NEPTUNE_SUBPROCESS_KILL_TIMEOUT = "NEPTUNE_SUBPROCESS_KILL_TIMEOUT"



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from neptune.new.run import Run


class Logger(object):
    def __init__(self, run: Run, attribute_name: str):
        self._run = run
        self._attribute_name = attribute_name

    def log(self, msg: str):
        self._run[self._attribute_name].log(msg)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import uuid
import unittest

from mock import patch, MagicMock, Mock
from bravado.exception import (
    HTTPNotFound,
    HTTPForbidden,
    HTTPConflict,
    HTTPBadRequest,
    HTTPUnprocessableEntity,
)
from bravado.testing.response_mocks import BravadoResponseMock

from neptune.management import (
    get_project_list,
    get_workspace_member_list,
    get_project_member_list,
    delete_project,
    create_project,
    add_project_member,
    remove_project_member,
    MemberRole,
)
from neptune.management.exceptions import (
    WorkspaceNotFound,
    ProjectNotFound,
    AccessRevokedOnDeletion,
    ProjectAlreadyExists,
    UserNotExistsOrWithoutAccess,
    UserAlreadyHasAccess,
    AccessRevokedOnMemberRemoval,
    UnsupportedValue,
)
from neptune.new.internal.backends.utils import verify_host_resolution
from neptune.new.internal.backends.hosted_client import (
    _get_token_client,  # pylint:disable=protected-access
    get_client_config,
    create_http_client_with_auth,
    create_backend_client,
    create_leaderboard_client,
    create_artifacts_client,
)
from tests.neptune.new.backend_test_mixin import BackendTestMixin

API_TOKEN = (
    "eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYXBwLnN0YWdlLm5lcHR1bmUubWwiLCJ"
    "hcGlfa2V5IjoiOTJhNzhiOWQtZTc3Ni00ODlhLWI5YzEtNzRkYmI1ZGVkMzAyIn0="
)


@patch("neptune.new.internal.backends.hosted_client.RequestsClient", new=MagicMock())
@patch(
    "neptune.new.internal.backends.hosted_client.NeptuneAuthenticator", new=MagicMock()
)
@patch("bravado.client.SwaggerClient.from_url")
@patch("platform.platform", new=lambda: "testPlatform")
@patch("platform.python_version", new=lambda: "3.9.test")
class TestHostedClient(unittest.TestCase, BackendTestMixin):
    def setUp(self) -> None:
        # Clear all LRU storage
        verify_host_resolution.cache_clear()
        _get_token_client.cache_clear()
        get_client_config.cache_clear()
        create_http_client_with_auth.cache_clear()
        create_backend_client.cache_clear()
        create_leaderboard_client.cache_clear()
        create_artifacts_client.cache_clear()

    def test_project_listing_empty(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        swagger_client.api.listProjects.return_value.response = BravadoResponseMock(
            result=Mock(entries=[])
        )

        # when:
        returned_projects = get_project_list(api_token=API_TOKEN)

        # then:
        self.assertEqual([], returned_projects)

    def test_project_listing(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        project1 = MagicMock(organizationName="org1")
        project1.name = "project1"
        project2 = MagicMock(organizationName="org2")
        project2.name = "project2"
        projects = Mock(entries=[project1, project2])
        swagger_client.api.listProjects.return_value.response = BravadoResponseMock(
            result=projects,
        )

        # when:
        returned_projects = get_project_list(api_token=API_TOKEN)

        # then:
        self.assertEqual(["org1/project1", "org2/project2"], returned_projects)

    def test_workspace_members(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        members = [
            Mock(role="member", registeredMemberInfo=Mock(username="tester1")),
            Mock(role="owner", registeredMemberInfo=Mock(username="tester2")),
        ]
        swagger_client.api.listOrganizationMembers.return_value.response = (
            BravadoResponseMock(
                result=members,
            )
        )

        # when:
        returned_members = get_workspace_member_list(name="org2", api_token=API_TOKEN)

        # then:
        self.assertEqual({"tester1": "member", "tester2": "admin"}, returned_members)

    def test_workspace_members_empty(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        members = []
        swagger_client.api.listOrganizationMembers.return_value.response = (
            BravadoResponseMock(
                result=members,
            )
        )

        # when:
        returned_members = get_workspace_member_list(name="org2", api_token=API_TOKEN)

        # then:
        self.assertEqual({}, returned_members)

    def test_workspace_members_not_exists(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.listOrganizationMembers.side_effect = HTTPNotFound(
            response=MagicMock()
        )

        # then:
        with self.assertRaises(WorkspaceNotFound):
            get_workspace_member_list(name="org2", api_token=API_TOKEN)

    def test_project_members(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        members = [
            Mock(role="member", registeredMemberInfo=Mock(username="tester1")),
            Mock(role="manager", registeredMemberInfo=Mock(username="tester2")),
            Mock(role="viewer", registeredMemberInfo=Mock(username="tester3")),
        ]
        swagger_client.api.listProjectMembers.return_value.response = (
            BravadoResponseMock(
                result=members,
            )
        )

        # when:
        returned_members = get_project_member_list(name="org/proj", api_token=API_TOKEN)

        # then:
        self.assertEqual(
            {"tester1": "contributor", "tester2": "owner", "tester3": "viewer"},
            returned_members,
        )

    def test_project_members_empty(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        members = []
        swagger_client.api.listProjectMembers.return_value.response = (
            BravadoResponseMock(
                result=members,
            )
        )

        # when:
        returned_members = get_project_member_list(name="org/proj", api_token=API_TOKEN)

        # then:
        self.assertEqual({}, returned_members)

    def test_project_members_not_exists(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.listProjectMembers.side_effect = HTTPNotFound(
            response=MagicMock()
        )

        # then:
        with self.assertRaises(ProjectNotFound):
            get_project_member_list(name="org/proj", api_token=API_TOKEN)

    def test_delete_project_not_exists(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.deleteProject.side_effect = HTTPNotFound(
            response=MagicMock()
        )

        # then:
        with self.assertRaises(ProjectNotFound):
            delete_project(name="org/proj", api_token=API_TOKEN)

    def test_delete_project_permissions(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.deleteProject.side_effect = HTTPForbidden(
            response=MagicMock()
        )

        # then:
        with self.assertRaises(AccessRevokedOnDeletion):
            delete_project(name="org/proj", api_token=API_TOKEN)

    def test_create_project_already_exists(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        organization = Mock(id=str(uuid.uuid4()))
        organization.name = "org"
        organizations = [organization]

        # when:
        swagger_client.api.listOrganizations.return_value.response = (
            BravadoResponseMock(
                result=organizations,
            )
        )
        swagger_client.api.createProject.side_effect = HTTPBadRequest(
            response=MagicMock(),
            swagger_result=MagicMock(
                code=None,
                errorType={"name": "validationError"},
                message=None,
                title="Validation Errors",
                type=None,
                validationErrors=[
                    {
                        "path": ["name"],
                        "errors": [{"errorCode": {"name": "ERR_NOT_UNIQUE"}}],
                    }
                ],
            ),
        )

        # then:
        with self.assertRaises(ProjectAlreadyExists):
            create_project(name="org/proj", key="PRJ", api_token=API_TOKEN)

    def test_create_project_unknown_visibility(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        organization = Mock(id=str(uuid.uuid4()))
        organization.name = "org"
        organizations = [organization]

        # when:
        swagger_client.api.listOrganizations.return_value.response = (
            BravadoResponseMock(
                result=organizations,
            )
        )

        with self.assertRaises(UnsupportedValue):
            create_project(
                name="org/proj",
                key="PRJ",
                visibility="unknown_value",
                api_token=API_TOKEN,
            )

    def test_create_project_no_workspace(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        organization = Mock(id=str(uuid.uuid4()))
        organization.name = "org"
        organizations = [organization]

        # when:
        swagger_client.api.listOrganizations.return_value.response = (
            BravadoResponseMock(
                result=organizations,
            )
        )

        # then:
        with self.assertRaises(WorkspaceNotFound):
            create_project(name="not_an_org/proj", key="PRJ", api_token=API_TOKEN)

    def test_create_project(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        organization = Mock(id=str(uuid.uuid4()))
        organization.name = "org"
        organizations = [organization]

        # and:
        project = Mock(organizationName="org")
        project.name = "proj"

        # when:
        swagger_client.api.listOrganizations.return_value.response = (
            BravadoResponseMock(
                result=organizations,
            )
        )
        swagger_client.api.createProject.return_value.response = BravadoResponseMock(
            result=project,
        )

        # then:
        self.assertEqual(
            "org/proj", create_project(name="org/proj", key="PRJ", api_token=API_TOKEN)
        )

    def test_add_project_member_project_not_found(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.addProjectMember.side_effect = HTTPNotFound(
            response=MagicMock()
        )

        # then:
        with self.assertRaises(ProjectNotFound):
            add_project_member(
                name="org/proj",
                username="tester",
                role=MemberRole.VIEWER,
                api_token=API_TOKEN,
            )

    def test_add_project_member_unknown_role(self, swagger_client_factory):
        _ = self._get_swagger_client_mock(swagger_client_factory)

        # then:
        with self.assertRaises(UnsupportedValue):
            add_project_member(
                name="org/proj",
                username="tester",
                role="unknown_role",
                api_token=API_TOKEN,
            )

    def test_add_project_member_member_without_access(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.addProjectMember.side_effect = HTTPConflict(
            response=MagicMock()
        )

        # then:
        with self.assertRaises(UserAlreadyHasAccess):
            add_project_member(
                name="org/proj",
                username="tester",
                role=MemberRole.VIEWER,
                api_token=API_TOKEN,
            )

    def test_remove_project_member_project_not_found(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.deleteProjectMember.side_effect = HTTPNotFound(
            response=MagicMock()
        )

        # then:
        with self.assertRaises(ProjectNotFound):
            remove_project_member(
                name="org/proj", username="tester", api_token=API_TOKEN
            )

    def test_remove_project_member_no_user(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.deleteProjectMember.side_effect = HTTPUnprocessableEntity(
            response=MagicMock()
        )

        # then:
        with self.assertRaises(UserNotExistsOrWithoutAccess):
            remove_project_member(
                name="org/proj", username="tester", api_token=API_TOKEN
            )

    def test_remove_project_member_permissions(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.deleteProjectMember.side_effect = HTTPForbidden(
            response=MagicMock()
        )

        # then:
        with self.assertRaises(AccessRevokedOnMemberRemoval):
            remove_project_member(
                name="org/proj", username="tester", api_token=API_TOKEN
            )



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import threading

from neptune.new.attribute_container import AttributeContainer
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.background_job import BackgroundJob
from neptune.new.internal.operation_processors.operation_processor import (
    OperationProcessor,
)
from neptune.new.internal.state import ContainerState

# backwards compatibility
# pylint: disable=unused-import
from neptune.new.attributes.attribute import Attribute
from neptune.new.attributes.namespace import (
    NamespaceBuilder,
    Namespace as NamespaceAttr,
)
from neptune.new.exceptions import (
    MetadataInconsistency,
    InactiveRunException,
    NeptunePossibleLegacyUsageException,
)
from neptune.new.handler import Handler
from neptune.new.internal.backends.api_model import AttributeType
from neptune.new.internal.operation import DeleteAttribute
from neptune.new.internal.run_structure import ContainerStructure as RunStructure
from neptune.new.internal.utils import (
    is_bool,
    is_float,
    is_float_like,
    is_int,
    is_string,
    is_string_like,
    verify_type,
    is_dict_like,
)
from neptune.new.internal.utils.paths import parse_path
from neptune.new.internal.value_to_attribute_visitor import ValueToAttributeVisitor
from neptune.new.types import Boolean, Integer
from neptune.new.types.atoms.datetime import Datetime
from neptune.new.types.atoms.float import Float
from neptune.new.types.atoms.string import String
from neptune.new.types.namespace import Namespace
from neptune.new.types.value import Value

RunState = ContainerState


class Run(AttributeContainer):
    """A Run in Neptune is a representation of all metadata that you log to Neptune.

    Beginning when you start a tracked run with `neptune.init()` and ending when the script finishes
    or when you explicitly stop the experiment with `.stop()`.

    You can log many ML metadata types, including:
        * metrics
        * losses
        * model weights
        * images
        * interactive charts
        * predictions
        * and much more

    Examples:
        >>> import neptune.new as neptune

        >>> # Create new experiment
        ... run = neptune.init('my_workspace/my_project')

        >>> # Log parameters
        ... params = {'max_epochs': 10, 'optimizer': 'Adam'}
        ... run['parameters'] = params

        >>> # Log metadata
        ... run['train/metric_name'].log()
        >>> run['predictions'].log(image)
        >>> run['model'].upload(path_to_model)

        >>> # Log whatever else you want
        ... pass

        >>> # Stop tracking and clean up
        ... run.stop()

    You may also want to check `Run docs page`_.

    .. _Run docs page:
       https://docs.neptune.ai/api-reference/run
    """

    last_run = None  # "static" instance of recently created Run

    LEGACY_METHODS = (
        "create_experiment",
        "send_metric",
        "log_metric",
        "send_text",
        "log_text",
        "send_image",
        "log_image",
        "send_artifact",
        "log_artifact",
        "delete_artifacts",
        "download_artifact",
        "download_sources",
        "download_artifacts",
        "reset_log",
        "get_parameters",
        "get_properties",
        "set_property",
        "remove_property",
        "get_hardware_utilization",
        "get_numeric_channels_values",
    )

    def __init__(
        self,
        _id: str,
        backend: NeptuneBackend,
        op_processor: OperationProcessor,
        background_job: BackgroundJob,
        lock: threading.RLock,
        workspace: str,
        project_name: str,
        short_id: str,
        project_id: str,
        monitoring_namespace: str = "monitoring",
    ):
        super().__init__(_id, backend, op_processor, background_job, lock, project_id)
        self._workspace = workspace
        self._project_name = project_name
        self._short_id = short_id
        self.monitoring_namespace = monitoring_namespace

        Run.last_run = self

    def get_run_url(self) -> str:
        """Returns the URL the run can be accessed with in the browser"""
        return self._backend.get_run_url(
            self._id, self._workspace, self._project_name, self._short_id
        )



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import socket
import unittest
import uuid
from unittest.mock import call

from bravado.exception import HTTPNotFound, HTTPPaymentRequired, HTTPUnprocessableEntity
from mock import MagicMock, patch
from packaging.version import Version

from neptune.new.exceptions import (
    CannotResolveHostname,
    FileUploadError,
    MetadataInconsistency,
    UnsupportedClientVersion,
    NeptuneLimitExceedException,
)
from neptune.new.internal.backends.hosted_neptune_backend import HostedNeptuneBackend
from neptune.new.internal.backends.hosted_client import (
    DEFAULT_REQUEST_KWARGS,
    _get_token_client,  # pylint:disable=protected-access
    get_client_config,
    create_http_client_with_auth,
    create_backend_client,
    create_leaderboard_client,
    create_artifacts_client,
)
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.operation import (
    AssignString,
    LogFloats,
    TrackFilesToArtifact,
    UploadFile,
    UploadFileContent,
)
from neptune.new.internal.backends.utils import verify_host_resolution
from neptune.new.internal.utils import base64_encode
from tests.neptune.new.backend_test_mixin import BackendTestMixin

API_TOKEN = (
    "eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYXBwLnN0YWdlLm5lcHR1bmUubWwiLCJ"
    "hcGlfa2V5IjoiOTJhNzhiOWQtZTc3Ni00ODlhLWI5YzEtNzRkYmI1ZGVkMzAyIn0="
)

credentials = Credentials.from_token(API_TOKEN)


@patch("neptune.new.internal.backends.hosted_client.RequestsClient", new=MagicMock())
@patch(
    "neptune.new.internal.backends.hosted_client.NeptuneAuthenticator", new=MagicMock()
)
@patch("bravado.client.SwaggerClient.from_url")
@patch("platform.platform", new=lambda: "testPlatform")
@patch("platform.python_version", new=lambda: "3.9.test")
class TestHostedNeptuneBackend(unittest.TestCase, BackendTestMixin):
    # pylint:disable=protected-access

    def setUp(self) -> None:
        # Clear all LRU storage
        verify_host_resolution.cache_clear()
        _get_token_client.cache_clear()
        get_client_config.cache_clear()
        create_http_client_with_auth.cache_clear()
        create_backend_client.cache_clear()
        create_leaderboard_client.cache_clear()
        create_artifacts_client.cache_clear()

    @patch("neptune.new.internal.backends.hosted_neptune_backend.upload_file_attribute")
    def test_execute_operations(self, upload_mock, swagger_client_factory):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        exp_uuid = str(uuid.uuid4())

        response_error = MagicMock()
        response_error.errorDescription = "error1"
        swagger_client.api.executeOperations().response().result = [response_error]
        swagger_client.api.executeOperations.reset_mock()
        upload_mock.return_value = FileUploadError("file1", "error2")
        some_text = "Some streamed text"
        some_binary = b"Some streamed binary"

        # when
        result = backend.execute_operations(
            run_id=exp_uuid,
            operations=[
                UploadFile(
                    path=["some", "files", "some_file"],
                    ext="",
                    file_path="path_to_file",
                ),
                UploadFileContent(
                    path=["some", "files", "some_text_stream"],
                    ext="txt",
                    file_content=base64_encode(some_text.encode("utf-8")),
                ),
                UploadFileContent(
                    path=["some", "files", "some_binary_stream"],
                    ext="bin",
                    file_content=base64_encode(some_binary),
                ),
                LogFloats(["images", "img1"], [LogFloats.ValueType(1, 2, 3)]),
                AssignString(["properties", "name"], "some text"),
                UploadFile(
                    path=["some", "other", "file.txt"],
                    ext="txt",
                    file_path="other/file/path.txt",
                ),
            ],
        )

        # than
        swagger_client.api.executeOperations.assert_called_once_with(
            **{
                "experimentId": str(exp_uuid),
                "operations": [
                    {
                        "path": "images/img1",
                        "logFloats": {
                            "entries": [
                                {"value": 1, "step": 2, "timestampMilliseconds": 3000}
                            ]
                        },
                    },
                    {"path": "properties/name", "assignString": {"value": "some text"}},
                ],
                **DEFAULT_REQUEST_KWARGS,
            }
        )

        upload_mock.assert_has_calls(
            [
                call(
                    swagger_client=backend.leaderboard_client,
                    run_id=exp_uuid,
                    attribute="some/other/file.txt",
                    source="other/file/path.txt",
                    ext="txt",
                ),
                call(
                    swagger_client=backend.leaderboard_client,
                    run_id=exp_uuid,
                    attribute="some/files/some_file",
                    source="path_to_file",
                    ext="",
                ),
                call(
                    swagger_client=backend.leaderboard_client,
                    run_id=exp_uuid,
                    attribute="some/files/some_text_stream",
                    source=some_text.encode("utf-8"),
                    ext="txt",
                ),
                call(
                    swagger_client=backend.leaderboard_client,
                    run_id=exp_uuid,
                    attribute="some/files/some_binary_stream",
                    source=some_binary,
                    ext="bin",
                ),
            ],
            any_order=True,
        )

        self.assertEqual(
            [
                FileUploadError("file1", "error2"),
                FileUploadError("file1", "error2"),
                FileUploadError("file1", "error2"),
                FileUploadError("file1", "error2"),
                MetadataInconsistency("error1"),
            ],
            result,
        )

    @patch("neptune.new.internal.backends.hosted_neptune_backend.upload_file_attribute")
    def test_upload_files_destination_path(self, upload_mock, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        exp_uuid = str(uuid.uuid4())

        # when
        backend.execute_operations(
            run_id=exp_uuid,
            operations=[
                UploadFile(
                    path=["some", "path", "1", "var"], ext="", file_path="/path/to/file"
                ),
                UploadFile(
                    path=["some", "path", "2", "var"],
                    ext="txt",
                    file_path="/some.file/with.dots.txt",
                ),
                UploadFile(
                    path=["some", "path", "3", "var"],
                    ext="jpeg",
                    file_path="/path/to/some_image.jpeg",
                ),
            ],
        )

        upload_mock.assert_has_calls(
            [
                call(
                    swagger_client=backend.leaderboard_client,
                    run_id=exp_uuid,
                    attribute="some/path/1/var",
                    source="/path/to/file",
                    ext="",
                ),
                call(
                    swagger_client=backend.leaderboard_client,
                    run_id=exp_uuid,
                    attribute="some/path/2/var",
                    source="/some.file/with.dots.txt",
                    ext="txt",
                ),
                call(
                    swagger_client=backend.leaderboard_client,
                    run_id=exp_uuid,
                    attribute="some/path/3/var",
                    source="/path/to/some_image.jpeg",
                    ext="jpeg",
                ),
            ],
            any_order=True,
        )

    @patch("neptune.new.internal.backends.hosted_neptune_backend.track_to_new_artifact")
    def test_track_to_new_artifact(
        self, track_to_new_artifact_mock, swagger_client_factory
    ):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        exp_id = str(uuid.uuid4())
        project_id = str(uuid.uuid4())

        response_error = MagicMock()
        response_error.errorDescription = "error1"
        swagger_client.api.executeOperations.return_value.response.return_value.result = [
            response_error
        ]
        swagger_client.api.getArtifactAttribute.side_effect = HTTPNotFound(
            response=MagicMock()
        )

        # when
        backend.execute_operations(
            run_id=exp_id,
            operations=[
                TrackFilesToArtifact(
                    path=["sub", "one"],
                    project_id=project_id,
                    entries=[("/path/to/file", "/path/to")],
                ),
                TrackFilesToArtifact(
                    path=["sub", "two"],
                    project_id=project_id,
                    entries=[("/path/to/file1", None), ("/path/to/file2", None)],
                ),
                TrackFilesToArtifact(
                    path=["sub", "three"],
                    project_id=project_id,
                    entries=[("/path/to/file1", None)],
                ),
                TrackFilesToArtifact(
                    path=["sub", "three"],
                    project_id=project_id,
                    entries=[("/path/to/file2", None)],
                ),
            ],
        )

        # then
        track_to_new_artifact_mock.assert_has_calls(
            [
                call(
                    swagger_client=swagger_client,
                    project_id=project_id,
                    path=["sub", "one"],
                    parent_identifier=str(exp_id),
                    entries=[("/path/to/file", "/path/to")],
                    default_request_params=DEFAULT_REQUEST_KWARGS,
                ),
                call(
                    swagger_client=swagger_client,
                    project_id=project_id,
                    path=["sub", "two"],
                    parent_identifier=str(exp_id),
                    entries=[("/path/to/file1", None), ("/path/to/file2", None)],
                    default_request_params=DEFAULT_REQUEST_KWARGS,
                ),
                call(
                    swagger_client=swagger_client,
                    project_id=project_id,
                    path=["sub", "three"],
                    parent_identifier=str(exp_id),
                    entries=[("/path/to/file1", None), ("/path/to/file2", None)],
                    default_request_params=DEFAULT_REQUEST_KWARGS,
                ),
            ],
            any_order=True,
        )

    @patch(
        "neptune.new.internal.backends.hosted_neptune_backend.track_to_existing_artifact"
    )
    def test_track_to_existing_artifact(
        self, track_to_existing_artifact_mock, swagger_client_factory
    ):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        exp_id = str(uuid.uuid4())
        project_id = str(uuid.uuid4())

        response_error = MagicMock()
        response_error.errorDescription = "error1"
        swagger_client.api.executeOperations.return_value.response.return_value.result = [
            response_error
        ]
        swagger_client.api.getArtifactAttribute.return_value.response.return_value.result.hash = (
            "dummyHash"
        )

        # when
        backend.execute_operations(
            run_id=exp_id,
            operations=[
                TrackFilesToArtifact(
                    path=["sub", "one"],
                    project_id=project_id,
                    entries=[("/path/to/file", "/path/to")],
                ),
                TrackFilesToArtifact(
                    path=["sub", "two"],
                    project_id=project_id,
                    entries=[("/path/to/file1", None), ("/path/to/file2", None)],
                ),
                TrackFilesToArtifact(
                    path=["sub", "three"],
                    project_id=project_id,
                    entries=[("/path/to/file1", None)],
                ),
                TrackFilesToArtifact(
                    path=["sub", "three"],
                    project_id=project_id,
                    entries=[("/path/to/file2", None)],
                ),
            ],
        )

        # then
        track_to_existing_artifact_mock.assert_has_calls(
            [
                call(
                    swagger_client=swagger_client,
                    project_id=project_id,
                    path=["sub", "one"],
                    artifact_hash="dummyHash",
                    parent_identifier=str(exp_id),
                    entries=[("/path/to/file", "/path/to")],
                    default_request_params=DEFAULT_REQUEST_KWARGS,
                ),
                call(
                    swagger_client=swagger_client,
                    project_id=project_id,
                    path=["sub", "two"],
                    artifact_hash="dummyHash",
                    parent_identifier=str(exp_id),
                    entries=[("/path/to/file1", None), ("/path/to/file2", None)],
                    default_request_params=DEFAULT_REQUEST_KWARGS,
                ),
                call(
                    swagger_client=swagger_client,
                    project_id=project_id,
                    path=["sub", "three"],
                    artifact_hash="dummyHash",
                    parent_identifier=str(exp_id),
                    entries=[("/path/to/file1", None), ("/path/to/file2", None)],
                    default_request_params=DEFAULT_REQUEST_KWARGS,
                ),
            ],
            any_order=True,
        )

    @patch(
        "neptune.new.internal.backends.hosted_client.neptune_client_version",
        Version("0.5.13"),
    )
    def test_min_compatible_version_ok(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, min_compatible="0.5.13")

        # expect
        HostedNeptuneBackend(credentials)

    @patch(
        "neptune.new.internal.backends.hosted_client.neptune_client_version",
        Version("0.5.13"),
    )
    def test_min_compatible_version_fail(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, min_compatible="0.5.14")

        # expect
        with self.assertRaises(UnsupportedClientVersion) as ex:
            HostedNeptuneBackend(credentials)

        self.assertTrue("Please install neptune-client>=0.5.14" in str(ex.exception))

    @patch(
        "neptune.new.internal.backends.hosted_client.neptune_client_version",
        Version("0.5.13"),
    )
    def test_max_compatible_version_ok(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, max_compatible="0.5.12")

        # expect
        HostedNeptuneBackend(credentials)

    @patch(
        "neptune.new.internal.backends.hosted_client.neptune_client_version",
        Version("0.5.13"),
    )
    def test_max_compatible_version_fail(self, swagger_client_factory):
        # given
        self._get_swagger_client_mock(swagger_client_factory, max_compatible="0.4.999")

        # expect
        with self.assertRaises(UnsupportedClientVersion) as ex:
            HostedNeptuneBackend(credentials)

        self.assertTrue("Please install neptune-client==0.4.0" in str(ex.exception))

    @patch("socket.gethostbyname")
    def test_cannot_resolve_host(self, gethostname_mock, _):
        # given
        gethostname_mock.side_effect = socket.gaierror

        # expect
        with self.assertRaises(CannotResolveHostname):
            HostedNeptuneBackend(credentials)

    def test_limit_exceed(self, swagger_client_factory):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        exp_uuid = str(uuid.uuid4())

        # when:
        error = MagicMock()
        error.json.return_value = {"title": "Maximum storage limit reached"}
        swagger_client.api.executeOperations.side_effect = HTTPPaymentRequired(
            response=error
        )

        # then:
        with self.assertRaises(NeptuneLimitExceedException):
            backend.execute_operations(
                run_id=exp_uuid,
                operations=[
                    LogFloats(["float1"], [LogFloats.ValueType(1, 2, 3)]),
                ],
            )

    def test_limit_exceed_legacy(self, swagger_client_factory):
        # given
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)
        backend = HostedNeptuneBackend(credentials)
        exp_uuid = str(uuid.uuid4())

        # when:
        error = MagicMock()
        error.json.return_value = {"title": "Monitoring hours not left"}
        swagger_client.api.executeOperations.side_effect = HTTPUnprocessableEntity(
            response=error
        )

        # then:
        with self.assertRaises(NeptuneLimitExceedException):
            backend.execute_operations(
                run_id=exp_uuid,
                operations=[
                    LogFloats(["float1"], [LogFloats.ValueType(1, 2, 3)]),
                ],
            )



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Union


from neptune.new.internal.operation import AssignString
from neptune.new.types.atoms.string import String as StringVal
from neptune.new.attributes.atoms.atom import Atom


class String(Atom):
    def assign(self, value: Union[StringVal, str], wait: bool = False):
        if not isinstance(value, StringVal):
            value = StringVal(value)

        with self._container.lock():
            self._enqueue_operation(AssignString(self._path, value.value), wait)

    def fetch(self) -> str:
        # pylint: disable=protected-access
        val = self._backend.get_string_attribute(self._container_id, self._path)
        return val.value



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

from neptune.management.internal.utils import normalize_project_name
from neptune.management.exceptions import (
    InvalidProjectName,
    ConflictingWorkspaceName,
    MissingWorkspaceName,
)


class TestManagementUtils(unittest.TestCase):
    def test_normalize_project_name(self):
        self.assertEqual(
            "jackie/sandbox", normalize_project_name(name="jackie/sandbox")
        )
        self.assertEqual(
            "jackie/sandbox", normalize_project_name(name="sandbox", workspace="jackie")
        )
        self.assertEqual(
            "jackie/sandbox",
            normalize_project_name(name="jackie/sandbox", workspace="jackie"),
        )

        with self.assertRaises(InvalidProjectName):
            normalize_project_name(name="nothing/else/matters")

        with self.assertRaises(MissingWorkspaceName):
            normalize_project_name(name="sandbox")

        with self.assertRaises(ConflictingWorkspaceName):
            normalize_project_name(name="jackie/sandbox", workspace="john")



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Remember to set environment values:
* NEPTUNE_API_TOKEN
* NEPTUNE_PROJECT
"""
import sys
from datetime import datetime

from requests import HTTPError

import neptune
from alpha_integration_dev.common_client_code import ClientFeatures
from neptune.exceptions import (
    DeleteArtifactUnsupportedInAlphaException,
    DownloadArtifactUnsupportedException,
    DownloadArtifactsUnsupportedException,
    DownloadSourcesException,
    FileNotFound,
)
from neptune.internal.api_clients.hosted_api_clients.hosted_alpha_leaderboard_api_client import (
    HostedAlphaLeaderboardApiClient,
)


def get_api_version(exp):
    backend = exp._backend
    if isinstance(backend, HostedAlphaLeaderboardApiClient):
        return 2
    return 1


class OldClientFeatures(ClientFeatures):
    def __init__(self):
        super().__init__()
        neptune.init()
        neptune.create_experiment(
            name="const project name",
            description="exp description",
            params=self.params,
            properties=self.properties,
            tags=["initial tag 1", "initial tag 2"],
            abort_callback=None,
            run_monitoring_thread=False,
            hostname="hostname value",
            # notebook_id='test1',  # TODO: Error 500 when wrong value
            upload_source_files="alpha_integration_dev/*.py",
        )

        exp = neptune.get_experiment()

        self._api_version = get_api_version(exp)

        properties = exp.get_properties()
        assert properties["init_text_property"] == "some text"
        assert properties["init_number property"] == "42"
        assert properties["init_list"] == "[1, 2, 3]"

        assert set(exp.get_tags()) == {"initial tag 1", "initial tag 2"}

        # download sources
        if self._api_version == 1:
            # old domain

            with self.with_check_if_file_appears("old_client.py.zip"):
                exp.download_sources("alpha_integration_dev/old_client.py")
            with self.with_check_if_file_appears("alpha_integration_dev.zip"):
                exp.download_sources("alpha_integration_dev")

            with self.with_assert_raises(FileNotFound):
                exp.download_sources("non_existing")
        else:
            # new api

            with self.with_check_if_file_appears("files.zip"):
                exp.download_sources()
            with self.with_assert_raises(DownloadSourcesException):
                exp.download_sources("whatever")
            with self.with_check_if_file_appears("file_set_sources/files.zip"):
                exp.download_sources(destination_dir="file_set_sources")

    def modify_tags(self):
        neptune.append_tags("tag1")
        neptune.append_tag(["tag2_to_remove", "tag3"])
        neptune.remove_tag("tag2_to_remove")
        neptune.remove_tag("tag4_remove_non_existing")

        exp = neptune.get_experiment()
        assert set(exp.get_tags()) == {"initial tag 1", "initial tag 2", "tag1", "tag3"}

    def modify_properties(self):
        neptune.set_property("prop", "some text")
        neptune.set_property("prop_number", 42)
        neptune.set_property("nested/prop", 42)
        neptune.set_property("prop_to_del", 42)
        neptune.set_property("prop_list", [1, 2, 3])
        with open(self.text_file_path, mode="r") as f:
            neptune.set_property("prop_IO", f)
        neptune.set_property("prop_datetime", datetime.now())
        neptune.remove_property("prop_to_del")

        exp = neptune.get_experiment()
        properties = exp.get_properties()
        assert properties["prop"] == "some text"
        assert properties["prop_number"] == "42"
        assert properties["nested/prop"] == "42"
        assert properties["prop_list"] == "[1, 2, 3]"
        assert "prop_to_del" not in properties
        assert (
            properties["prop_IO"]
            == "<_io.TextIOWrapper name='alpha_integration_dev/data/text.txt'"
            " mode='r' encoding='UTF-8'>"
        )
        print(f"Properties: {properties}")

    def log_std(self):
        print("stdout text1")
        print("stdout text2")
        print("stderr text1", file=sys.stderr)
        print("stderr text2", file=sys.stderr)

    def log_series(self):
        # floats
        neptune.log_metric("m1", 1)
        neptune.log_metric("m1", 2)
        neptune.log_metric("m1", 3)
        neptune.log_metric("m1", 2)
        neptune.log_metric("nested/m1", 1)

        # texts
        neptune.log_text("m2", "a")
        neptune.log_text("m2", "b")
        neptune.log_text("m2", "c")

        # images
        # `image_name` and `description` will be lost
        neptune.log_image("g_img", self.img_path, image_name="name", description="desc")
        neptune.log_image("g_img", self.img_path)

        # see what we've logged
        logs = neptune.get_experiment().get_logs()
        print(f"Logs: {logs}")

    def handle_files_and_images(self):
        # image
        # `image_name` and `description` will be lost (`send_image` the same as `log_image`)
        neptune.send_image("image", self.img_path, name="name", description="desc")

        # artifact with default dest
        neptune.send_artifact(self.text_file_path)
        exp = neptune.get_experiment()
        with self.with_check_if_file_appears("text.txt"):
            exp.download_artifact("text.txt")
        with self.with_check_if_file_appears("custom_dest/text.txt"):
            exp.download_artifact("text.txt", "custom_dest")

        # artifact with custom dest
        neptune.send_artifact(self.text_file_path, destination="something.txt")
        exp = neptune.get_experiment()
        with self.with_check_if_file_appears("something.txt"):
            exp.download_artifact("something.txt")
        with self.with_check_if_file_appears("custom_dest/something.txt"):
            exp.download_artifact("something.txt", "custom_dest")

        # destination dirs
        neptune.log_artifact(self.text_file_path, destination="dir/text file artifact")
        neptune.log_artifact(self.text_file_path, destination="dir/artifact_to_delete")

        # deleting
        neptune.delete_artifacts("dir/artifact_to_delete")

        # streams
        with open(self.text_file_path, mode="r") as f:
            neptune.send_artifact(f, destination="file stream.txt")

    def handle_directories(self):
        exp = neptune.get_experiment()

        # download_artifacts
        neptune.send_artifact(self.data_dir)
        if self._api_version == 1:
            with self.with_check_if_file_appears("output.zip"):
                exp.download_artifacts()
        else:
            with self.with_assert_raises(DownloadArtifactsUnsupportedException):
                exp.download_artifacts()

        # create some nested artifacts
        neptune.log_artifact(self.img_path, destination="main dir/sub dir/art1")
        neptune.log_artifact(self.img_path, destination="main dir/sub dir/art2")
        neptune.log_artifact(self.img_path, destination="main dir/sub dir/art3")

        # downloading artifact - download_artifact
        # non existing artifact
        if self._api_version == 1:
            with self.with_assert_raises(FileNotFound):
                exp.download_artifact("main dir/sub dir/art100")
        else:
            with self.with_assert_raises(DownloadArtifactUnsupportedException):
                exp.download_artifact("main dir/sub dir/art100")
        # artifact directories
        if self._api_version == 1:
            with self.with_assert_raises(HTTPError):
                exp.download_artifact("main dir/sub dir")
        else:
            with self.with_assert_raises(DownloadArtifactUnsupportedException):
                exp.download_artifact("main dir/sub dir")

        # deleting artifacts
        neptune.delete_artifacts("main dir/sub dir/art1")

        # delete non existing artifact
        if self._api_version == 1:
            neptune.delete_artifacts("main dir/sub dir/art100")
        else:
            with self.with_assert_raises(DeleteArtifactUnsupportedInAlphaException):
                neptune.delete_artifacts("main dir/sub dir/art100")

        # delete dir
        if self._api_version == 1:
            neptune.delete_artifacts("main dir/sub dir")
        else:
            with self.with_assert_raises(DeleteArtifactUnsupportedInAlphaException):
                neptune.delete_artifacts("main dir/sub dir")

    def finalize(self):
        pass


if __name__ == "__main__":
    OldClientFeatures().run()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time
import unittest

import jwt
from mock import MagicMock, patch
from oauthlib.oauth2 import TokenExpiredError

from neptune.oauth import (
    NeptuneAuth,
    NeptuneAuthenticator,
    _no_token_updater,
    _decoding_options,
)
from tests.neptune.http_objects_factory import a_request
from tests.neptune.oauth_objects_factory import SECRET, a_refresh_token, an_access_token


class TestNeptuneAuth(unittest.TestCase):
    def setUp(self):
        super(TestNeptuneAuth, self).setUp()

        self.session = MagicMock()
        self.session.token = dict()
        self.neptune_auth = NeptuneAuth(lambda: self.session)
        self.neptune_auth.token_expires_at = time.time() + 60
        self.request = a_request()

        self.url, self.method, self.body, self.headers = (
            self.request.url,
            self.request.method,
            self.request.body,
            self.request.headers,
        )

        self.updated_url, self.updated_headers, self.updated_body = (
            a_request().url,
            a_request().headers,
            a_request().body,
        )

    def test_add_valid_token(self):
        # given
        # pylint: disable=protected-access
        self.session._client.add_token.return_value = (
            self.updated_url,
            self.updated_headers,
            self.updated_body,
        )

        # when
        updated_request = self.neptune_auth(self.request)

        # then
        # pylint: disable=protected-access
        self.session._client.add_token.assert_called_once_with(
            self.url, http_method=self.method, body=self.body, headers=self.headers
        )

        # and
        self.assertEqual(self.updated_url, updated_request.url)
        self.assertEqual(self.updated_headers, updated_request.headers)
        self.assertEqual(self.updated_body, updated_request.body)

    def test_refresh_token_and_add(self):
        # given
        # pylint: disable=protected-access
        self.session._client.add_token.side_effect = [
            TokenExpiredError,
            (self.updated_url, self.updated_headers, self.updated_body),
        ]

        # when
        updated_request = self.neptune_auth(self.request)

        # then
        # pylint: disable=protected-access
        self.session._client.add_token.assert_called_with(
            self.url, http_method=self.method, body=self.body, headers=self.headers
        )

        # and
        self.assertEqual(self.updated_url, updated_request.url)
        self.assertEqual(self.updated_headers, updated_request.headers)
        self.assertEqual(self.updated_body, updated_request.body)


class TestNeptuneAuthenticator(unittest.TestCase):
    @patch("neptune.oauth.OAuth2Session")
    @patch("neptune.oauth.time")
    def test_apply_oauth2_session_to_request(self, time_mock, session_mock):
        # given
        api_token = MagicMock()
        backend_client = MagicMock()

        auth_tokens = MagicMock()
        auth_tokens.accessToken = an_access_token()
        auth_tokens.refreshToken = a_refresh_token()
        decoded_access_token = jwt.decode(
            auth_tokens.accessToken, SECRET, options=_decoding_options
        )

        backend_client.api.exchangeApiToken(
            X_Neptune_Api_Token=api_token
        ).response().result = auth_tokens

        # and
        now = time.time()
        time_mock.time.return_value = now

        # and
        session = MagicMock()
        session_mock.return_value = session
        session.token = dict()

        # and
        neptune_authenticator = NeptuneAuthenticator(
            api_token, backend_client, False, None
        )
        request = a_request()

        # when
        updated_request = neptune_authenticator.apply(request)

        # then
        expected_token = {
            "access_token": auth_tokens.accessToken,
            "refresh_token": auth_tokens.refreshToken,
            "expires_in": decoded_access_token["exp"] - now,
        }

        expected_auto_refresh_url = "{realm_url}/protocol/openid-connect/token".format(
            realm_url=decoded_access_token["iss"]
        )

        session_mock.assert_called_once_with(
            client_id=decoded_access_token["azp"],
            token=expected_token,
            auto_refresh_url=expected_auto_refresh_url,
            auto_refresh_kwargs={"client_id": decoded_access_token["azp"]},
            token_updater=_no_token_updater,
        )

        # and
        self.assertEqual(session, updated_request.auth.session)


if __name__ == "__main__":
    unittest.main()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import re
from typing import List, Optional, Dict, Iterable, Tuple, Any

from bravado.exception import HTTPNotFound, HTTPUnprocessableEntity, HTTPPaymentRequired
from simplejson import JSONDecodeError

from neptune.new.internal.artifacts.types import ArtifactFileData
from neptune.patterns import PROJECT_QUALIFIED_NAME_PATTERN
from neptune.new.exceptions import (
    ClientHttpError,
    FetchAttributeNotFoundException,
    RunNotFound,
    RunUUIDNotFound,
    InternalClientError,
    MetadataInconsistency,
    NeptuneException,
    NeptuneLegacyProjectException,
    ProjectNotFound,
    ProjectNameCollision,
    NeptuneLimitExceedException,
    ArtifactNotFoundException,
    NeptuneFeaturesNotAvailableException,
)
from neptune.new.internal.backends.api_model import (
    ApiRun,
    ArtifactAttribute,
    Attribute,
    AttributeType,
    AttributeWithProperties,
    BoolAttribute,
    DatetimeAttribute,
    FileAttribute,
    FloatAttribute,
    FloatPointValue,
    FloatSeriesAttribute,
    FloatSeriesValues,
    ImageSeriesValues,
    IntAttribute,
    LeaderboardEntry,
    Project,
    StringAttribute,
    StringPointValue,
    StringSeriesAttribute,
    StringSeriesValues,
    StringSetAttribute,
    Workspace,
)
from neptune.new.internal.backends.hosted_file_operations import (
    download_file_attribute,
    download_file_set_attribute,
    download_image_series_element,
    upload_file_attribute,
    upload_file_set_attribute,
)
from neptune.new.internal.backends.hosted_artifact_operations import (
    track_to_new_artifact,
    track_to_existing_artifact,
)
from neptune.new.internal.backends.neptune_backend import NeptuneBackend
from neptune.new.internal.backends.operation_api_name_visitor import (
    OperationApiNameVisitor,
)
from neptune.new.internal.backends.operation_api_object_converter import (
    OperationApiObjectConverter,
)
from neptune.new.internal.backends.operations_preprocessor import OperationsPreprocessor
from neptune.new.internal.backends.utils import (
    with_api_exceptions_handler,
    MissingApiClient,
    OptionalFeatures,
)
from neptune.new.internal.backends.hosted_client import (
    DEFAULT_REQUEST_KWARGS,
    create_http_client_with_auth,
    create_backend_client,
    create_leaderboard_client,
    create_artifacts_client,
)
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.operation import (
    Operation,
    TrackFilesToArtifact,
    UploadFile,
    UploadFileContent,
    UploadFileSet,
)
from neptune.new.internal.utils import verify_type, base64_decode
from neptune.new.internal.utils.generic_attribute_mapper import (
    map_attribute_result_to_value,
)
from neptune.new.internal.utils.paths import path_to_str
from neptune.new.internal.backends.utils import build_operation_url, ssl_verify
from neptune.new.internal.websockets.websockets_factory import WebsocketsFactory
from neptune.new.types.atoms import GitRef
from neptune.new.version import version as neptune_client_version

_logger = logging.getLogger(__name__)


class HostedNeptuneBackend(NeptuneBackend):
    def __init__(
        self, credentials: Credentials, proxies: Optional[Dict[str, str]] = None
    ):
        self.credentials = credentials
        self.proxies = proxies
        self.missing_features = []

        self._http_client, self._client_config = create_http_client_with_auth(
            credentials=credentials, ssl_verify=ssl_verify(), proxies=proxies
        )

        self.backend_client = create_backend_client(
            self._client_config, self._http_client
        )
        self.leaderboard_client = create_leaderboard_client(
            self._client_config, self._http_client
        )

        try:
            self.artifacts_client = create_artifacts_client(
                self._client_config, self._http_client
            )
        except JSONDecodeError:
            # thanks for nice error handling, bravado
            self.artifacts_client = MissingApiClient(self)
            self.missing_features.append(OptionalFeatures.ARTIFACTS)

    def verify_feature_available(self, feature_name: str):
        if feature_name in self.missing_features:
            raise NeptuneFeaturesNotAvailableException(self.missing_features)

    def get_display_address(self) -> str:
        return self._client_config.display_url

    def websockets_factory(
        self, project_id: str, run_id: str
    ) -> Optional[WebsocketsFactory]:
        base_url = re.sub(r"^http", "ws", self._client_config.api_url)
        return WebsocketsFactory(
            url=build_operation_url(
                base_url, f"/api/notifications/v1/runs/{project_id}/{run_id}/signal"
            ),
            session=self._http_client.authenticator.auth.session,
            proxies=self.proxies,
        )

    def close(self) -> None:
        pass

    @with_api_exceptions_handler
    def get_project(self, project_id: str) -> Project:
        verify_type("project_id", project_id, str)

        project_spec = re.search(PROJECT_QUALIFIED_NAME_PATTERN, project_id)
        workspace, name = project_spec["workspace"], project_spec["project"]

        try:
            if not workspace:
                available_projects = list(
                    filter(
                        lambda p: p.name == name,
                        self.get_available_projects(search_term=name),
                    )
                )

                if len(available_projects) == 1:
                    project = available_projects[0]
                    project_id = f"{project.workspace}/{project.name}"
                elif len(available_projects) > 1:
                    raise ProjectNameCollision(
                        project_id=project_id, available_projects=available_projects
                    )
                else:
                    raise ProjectNotFound(
                        project_id=project_id,
                        available_projects=self.get_available_projects(),
                        available_workspaces=self.get_available_workspaces(),
                    )

            response = self.backend_client.api.getProject(
                projectIdentifier=project_id,
                **DEFAULT_REQUEST_KWARGS,
            ).response()
            project = response.result
            project_version = project.version if hasattr(project, "version") else 1
            if project_version < 2:
                raise NeptuneLegacyProjectException(project_id)
            return Project(project.id, project.name, project.organizationName)
        except HTTPNotFound:
            raise ProjectNotFound(
                project_id,
                available_projects=self.get_available_projects(workspace_id=workspace),
                available_workspaces=list()
                if workspace
                else self.get_available_workspaces(),
            )

    @with_api_exceptions_handler
    def get_available_projects(
        self, workspace_id: Optional[str] = None, search_term: Optional[str] = None
    ) -> List[Project]:
        try:
            response = self.backend_client.api.listProjects(
                limit=5,
                organizationIdentifier=workspace_id,
                searchTerm=search_term,
                sortBy=["lastViewed"],
                sortDirection=["descending"],
                userRelation="memberOrHigher",
                **DEFAULT_REQUEST_KWARGS,
            ).response()
            projects = response.result.entries
            return list(
                map(
                    lambda project: Project(
                        project.id, project.name, project.organizationName
                    ),
                    projects,
                )
            )
        except HTTPNotFound:
            return []

    @with_api_exceptions_handler
    def get_available_workspaces(self) -> List[Workspace]:
        try:
            response = self.backend_client.api.listOrganizations(
                **DEFAULT_REQUEST_KWARGS,
            ).response()
            workspaces = response.result
            return list(
                map(
                    lambda workspace: Workspace(_id=workspace.id, name=workspace.name),
                    workspaces,
                )
            )
        except HTTPNotFound:
            return []

    @with_api_exceptions_handler
    def get_run(self, run_id: str):
        try:
            run = (
                self.leaderboard_client.api.getExperiment(
                    experimentId=run_id,
                    **DEFAULT_REQUEST_KWARGS,
                )
                .response()
                .result
            )
            return ApiRun(
                run.id, run.shortId, run.organizationName, run.projectName, run.trashed
            )
        except HTTPNotFound:
            raise RunNotFound(run_id)

    @with_api_exceptions_handler
    def create_run(
        self,
        project_id: str,
        git_ref: Optional[GitRef] = None,
        custom_run_id: Optional[str] = None,
        notebook_id: Optional[str] = None,
        checkpoint_id: Optional[str] = None,
    ) -> ApiRun:
        verify_type("project_id", project_id, str)

        git_info = (
            {
                "commit": {
                    "commitId": git_ref.commit_id,
                    "message": git_ref.message,
                    "authorName": git_ref.author_name,
                    "authorEmail": git_ref.author_email,
                    "commitDate": git_ref.commit_date,
                },
                "repositoryDirty": git_ref.dirty,
                "currentBranch": git_ref.branch,
                "remotes": git_ref.remotes,
            }
            if git_ref
            else None
        )

        params = {
            "projectIdentifier": project_id,
            "cliVersion": str(neptune_client_version),
            "gitInfo": git_info,
            "customId": custom_run_id,
        }

        if notebook_id is not None and checkpoint_id is not None:
            params["notebookId"] = notebook_id if notebook_id is not None else None
            params["checkpointId"] = (
                checkpoint_id if checkpoint_id is not None else None
            )

        kwargs = {
            "experimentCreationParams": params,
            "X-Neptune-CliVersion": str(neptune_client_version),
            **DEFAULT_REQUEST_KWARGS,
        }

        try:
            run = (
                self.leaderboard_client.api.createExperiment(**kwargs).response().result
            )
            return ApiRun(
                run.id, run.shortId, run.organizationName, run.projectName, run.trashed
            )
        except HTTPNotFound:
            raise ProjectNotFound(project_id=project_id)

    @with_api_exceptions_handler
    def create_checkpoint(self, notebook_id: str, jupyter_path: str) -> Optional[str]:
        try:
            return (
                self.leaderboard_client.api.createEmptyCheckpoint(
                    notebookId=notebook_id,
                    checkpoint={"path": jupyter_path},
                    **DEFAULT_REQUEST_KWARGS,
                )
                .response()
                .result.id
            )
        except HTTPNotFound:
            return None

    @with_api_exceptions_handler
    def ping_run(self, run_id: str):
        request_kwargs = {
            "_request_options": {
                "timeout": 10,
                "connect_timeout": 10,
            }
        }
        try:
            self.leaderboard_client.api.ping(
                experimentId=run_id,
                **request_kwargs,
            ).response().result
        except HTTPNotFound:
            raise RunUUIDNotFound(run_id)

    def execute_operations(
        self, run_id: str, operations: List[Operation]
    ) -> List[NeptuneException]:
        errors = []

        operations_preprocessor = OperationsPreprocessor()
        operations_preprocessor.process(operations)
        errors.extend(operations_preprocessor.get_errors())

        upload_operations, artifact_operations, other_operations = [], [], []

        for op in operations_preprocessor.get_operations():
            if isinstance(op, (UploadFile, UploadFileContent, UploadFileSet)):
                upload_operations.append(op)
            elif isinstance(op, TrackFilesToArtifact):
                artifact_operations.append(op)
            else:
                other_operations.append(op)

        if artifact_operations:
            self.verify_feature_available(OptionalFeatures.ARTIFACTS)

        # Upload operations should be done first since they are idempotent
        errors.extend(
            self._execute_upload_operations_with_400_retry(
                run_id=run_id, upload_operations=upload_operations
            )
        )

        (
            artifact_operations_errors,
            assign_artifact_operations,
        ) = self._execute_artifact_operations(
            run_id=run_id, artifact_operations=artifact_operations
        )

        errors.extend(artifact_operations_errors)
        other_operations.extend(assign_artifact_operations)

        if other_operations:
            errors.extend(self._execute_operations(run_id, other_operations))

        return errors

    def _execute_upload_operations(
        self, run_id: str, upload_operations: List[Operation]
    ) -> List[NeptuneException]:
        errors = list()

        for op in upload_operations:
            if isinstance(op, UploadFile):
                error = upload_file_attribute(
                    swagger_client=self.leaderboard_client,
                    run_id=run_id,
                    attribute=path_to_str(op.path),
                    source=op.file_path,
                    ext=op.ext,
                )
                if error:
                    errors.append(error)
            elif isinstance(op, UploadFileContent):
                error = upload_file_attribute(
                    swagger_client=self.leaderboard_client,
                    run_id=run_id,
                    attribute=path_to_str(op.path),
                    source=base64_decode(op.file_content),
                    ext=op.ext,
                )
                if error:
                    errors.append(error)
            elif isinstance(op, UploadFileSet):
                error = upload_file_set_attribute(
                    swagger_client=self.leaderboard_client,
                    run_id=run_id,
                    attribute=path_to_str(op.path),
                    file_globs=op.file_globs,
                    reset=op.reset,
                )
                if error:
                    errors.append(error)
            else:
                raise InternalClientError("Upload operation in neither File or FileSet")

        return errors

    def _execute_upload_operations_with_400_retry(
        self, run_id: str, upload_operations: List[Operation]
    ) -> List[NeptuneException]:
        while True:
            try:
                return self._execute_upload_operations(run_id, upload_operations)
            except ClientHttpError as ex:
                if "Length of stream does not match given range" not in ex.response:
                    raise ex

    @with_api_exceptions_handler
    def _execute_artifact_operations(
        self, run_id: str, artifact_operations: List[TrackFilesToArtifact]
    ) -> Tuple[List[Optional[NeptuneException]], List[Optional[Operation]]]:
        errors = list()
        assign_operations = list()

        for op in artifact_operations:
            try:
                artifact_hash = self.get_artifact_attribute(run_id, op.path).hash
            except FetchAttributeNotFoundException:
                artifact_hash = None

            try:
                if artifact_hash is None:
                    assign_operation = track_to_new_artifact(
                        swagger_client=self.artifacts_client,
                        project_id=op.project_id,
                        path=op.path,
                        parent_identifier=run_id,
                        entries=op.entries,
                        default_request_params=DEFAULT_REQUEST_KWARGS,
                    )
                else:
                    assign_operation = track_to_existing_artifact(
                        swagger_client=self.artifacts_client,
                        project_id=op.project_id,
                        path=op.path,
                        artifact_hash=artifact_hash,
                        parent_identifier=run_id,
                        entries=op.entries,
                        default_request_params=DEFAULT_REQUEST_KWARGS,
                    )

                if assign_operation:
                    assign_operations.append(assign_operation)
            except NeptuneException as error:
                errors.append(error)

        return errors, assign_operations

    @with_api_exceptions_handler
    def _execute_operations(
        self, run_id: str, operations: List[Operation]
    ) -> List[MetadataInconsistency]:
        kwargs = {
            "experimentId": run_id,
            "operations": [
                {
                    "path": path_to_str(op.path),
                    OperationApiNameVisitor()
                    .visit(op): OperationApiObjectConverter()
                    .convert(op),
                }
                for op in operations
            ],
            **DEFAULT_REQUEST_KWARGS,
        }

        try:
            result = (
                self.leaderboard_client.api.executeOperations(**kwargs)
                .response()
                .result
            )
            return [MetadataInconsistency(err.errorDescription) for err in result]
        except HTTPNotFound as e:
            raise RunUUIDNotFound(run_id=run_id) from e
        except (HTTPPaymentRequired, HTTPUnprocessableEntity) as e:
            raise NeptuneLimitExceedException(
                reason=e.response.json().get("title", "Unknown reason")
            ) from e

    @with_api_exceptions_handler
    def get_attributes(self, run_id: str) -> List[Attribute]:
        def to_attribute(attr) -> Attribute:
            return Attribute(attr.name, AttributeType(attr.type))

        params = {
            "experimentId": run_id,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            run = (
                self.leaderboard_client.api.getExperimentAttributes(**params)
                .response()
                .result
            )

            attribute_type_names = [at.value for at in AttributeType]
            accepted_attributes = [
                attr for attr in run.attributes if attr.type in attribute_type_names
            ]

            # Notify about ignored attrs
            ignored_attributes = set(attr.type for attr in run.attributes) - set(
                attr.type for attr in accepted_attributes
            )
            if ignored_attributes:
                _logger.warning(
                    "Ignored following attributes (unknown type): %s.\n"
                    "Try to upgrade `neptune-client.",
                    ignored_attributes,
                )

            return [
                to_attribute(attr)
                for attr in accepted_attributes
                if attr.type in attribute_type_names
            ]
        except HTTPNotFound:
            raise RunUUIDNotFound(run_id=run_id)

    def download_file_series_by_index(
        self, run_id: str, path: List[str], index: int, destination: str
    ):
        try:
            download_image_series_element(
                swagger_client=self.leaderboard_client,
                run_id=run_id,
                attribute=path_to_str(path),
                index=index,
                destination=destination,
            )
        except ClientHttpError as e:
            if e.status == HTTPNotFound.status_code:
                raise FetchAttributeNotFoundException(path_to_str(path))
            else:
                raise

    def download_file(
        self, run_id: str, path: List[str], destination: Optional[str] = None
    ):
        try:
            download_file_attribute(
                swagger_client=self.leaderboard_client,
                run_id=run_id,
                attribute=path_to_str(path),
                destination=destination,
            )
        except ClientHttpError as e:
            if e.status == HTTPNotFound.status_code:
                raise FetchAttributeNotFoundException(path_to_str(path))
            else:
                raise

    def download_file_set(
        self, run_id: str, path: List[str], destination: Optional[str] = None
    ):
        download_request = self._get_file_set_download_request(run_id, path)
        try:
            download_file_set_attribute(
                swagger_client=self.leaderboard_client,
                download_id=download_request.id,
                destination=destination,
            )
        except ClientHttpError as e:
            if e.status == HTTPNotFound.status_code:
                raise FetchAttributeNotFoundException(path_to_str(path))
            else:
                raise

    @with_api_exceptions_handler
    def get_float_attribute(self, run_id: str, path: List[str]) -> FloatAttribute:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getFloatAttribute(**params)
                .response()
                .result
            )
            return FloatAttribute(result.value)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_int_attribute(self, run_id: str, path: List[str]) -> IntAttribute:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getIntAttribute(**params).response().result
            )
            return IntAttribute(result.value)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_bool_attribute(self, run_id: str, path: List[str]) -> BoolAttribute:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getBoolAttribute(**params).response().result
            )
            return BoolAttribute(result.value)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_file_attribute(self, run_id: str, path: List[str]) -> FileAttribute:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getFileAttribute(**params).response().result
            )
            return FileAttribute(name=result.name, ext=result.ext, size=result.size)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_attribute(self, run_id: str, path: List[str]) -> StringAttribute:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getStringAttribute(**params)
                .response()
                .result
            )
            return StringAttribute(result.value)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_datetime_attribute(self, run_id: str, path: List[str]) -> DatetimeAttribute:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getDatetimeAttribute(**params)
                .response()
                .result
            )
            return DatetimeAttribute(result.value)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_artifact_attribute(self, run_id: str, path: List[str]) -> ArtifactAttribute:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getArtifactAttribute(**params)
                .response()
                .result
            )
            return ArtifactAttribute(hash=result.hash)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def list_artifact_files(
        self, project_id: str, artifact_hash: str
    ) -> List[ArtifactFileData]:
        params = {
            "projectIdentifier": project_id,
            "hash": artifact_hash,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.artifacts_client.api.listArtifactFiles(**params).response().result
            )
            return [ArtifactFileData.from_dto(a) for a in result.files]
        except HTTPNotFound:
            raise ArtifactNotFoundException(artifact_hash)

    @with_api_exceptions_handler
    def get_float_series_attribute(
        self, run_id: str, path: List[str]
    ) -> FloatSeriesAttribute:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getFloatSeriesAttribute(**params)
                .response()
                .result
            )
            return FloatSeriesAttribute(result.last)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_series_attribute(
        self, run_id: str, path: List[str]
    ) -> StringSeriesAttribute:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getStringSeriesAttribute(**params)
                .response()
                .result
            )
            return StringSeriesAttribute(result.last)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_set_attribute(
        self, run_id: str, path: List[str]
    ) -> StringSetAttribute:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getStringSetAttribute(**params)
                .response()
                .result
            )
            return StringSetAttribute(set(result.values))
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_image_series_values(
        self, run_id: str, path: List[str], offset: int, limit: int
    ) -> ImageSeriesValues:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            "limit": limit,
            "offset": offset,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getImageSeriesValues(**params)
                .response()
                .result
            )
            return ImageSeriesValues(result.totalItemCount)
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_string_series_values(
        self, run_id: str, path: List[str], offset: int, limit: int
    ) -> StringSeriesValues:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            "limit": limit,
            "offset": offset,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getStringSeriesValues(**params)
                .response()
                .result
            )
            return StringSeriesValues(
                result.totalItemCount,
                [
                    StringPointValue(v.timestampMillis, v.step, v.value)
                    for v in result.values
                ],
            )
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_float_series_values(
        self, run_id: str, path: List[str], offset: int, limit: int
    ) -> FloatSeriesValues:
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            "limit": limit,
            "offset": offset,
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            result = (
                self.leaderboard_client.api.getFloatSeriesValues(**params)
                .response()
                .result
            )
            return FloatSeriesValues(
                result.totalItemCount,
                [
                    FloatPointValue(v.timestampMillis, v.step, v.value)
                    for v in result.values
                ],
            )
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def fetch_atom_attribute_values(
        self, run_id: str, path: List[str]
    ) -> List[Tuple[str, AttributeType, Any]]:
        params = {
            "experimentId": run_id,
        }
        try:
            namespace_prefix = path_to_str(path)
            if namespace_prefix:
                # don't want to catch "ns/attribute/other" while looking for "ns/attr"
                namespace_prefix += "/"
            result = (
                self.leaderboard_client.api.getExperimentAttributes(**params)
                .response()
                .result
            )
            return [
                (attr.name, attr.type, map_attribute_result_to_value(attr))
                for attr in result.attributes
                if attr.name.startswith(namespace_prefix)
            ]
        except HTTPNotFound:
            raise RunUUIDNotFound(run_id)

    @with_api_exceptions_handler
    def _get_file_set_download_request(self, run_id: str, path: List[str]):
        params = {
            "experimentId": run_id,
            "attribute": path_to_str(path),
            **DEFAULT_REQUEST_KWARGS,
        }
        try:
            return (
                self.leaderboard_client.api.prepareForDownloadFileSetAttributeZip(
                    **params
                )
                .response()
                .result
            )
        except HTTPNotFound:
            raise FetchAttributeNotFoundException(path_to_str(path))

    @with_api_exceptions_handler
    def get_leaderboard(
        self,
        project_id: str,
        _id: Optional[Iterable[str]] = None,
        state: Optional[Iterable[str]] = None,
        owner: Optional[Iterable[str]] = None,
        tags: Optional[Iterable[str]] = None,
    ) -> List[LeaderboardEntry]:
        def get_portion(limit, offset):
            return (
                self.leaderboard_client.api.getLeaderboard(
                    projectIdentifier=project_id,
                    shortId=_id,
                    state=state,
                    owner=owner,
                    tags=tags,
                    tagsMode="and",
                    sortBy=["shortId"],
                    sortFieldType=["string"],
                    sortDirection=["ascending"],
                    limit=limit,
                    offset=offset,
                    **DEFAULT_REQUEST_KWARGS,
                )
                .response()
                .result.entries
            )

        def to_leaderboard_entry(entry) -> LeaderboardEntry:
            supported_attribute_types = {item.value for item in AttributeType}
            attributes: List[AttributeWithProperties] = []
            for attr in entry.attributes:
                if attr.type in supported_attribute_types:
                    properties = attr.__getitem__("{}Properties".format(attr.type))
                    attributes.append(
                        AttributeWithProperties(
                            attr.name, AttributeType(attr.type), properties
                        )
                    )
            return LeaderboardEntry(entry.id, attributes)

        try:
            return [
                to_leaderboard_entry(e)
                for e in self._get_all_items(get_portion, step=100)
            ]
        except HTTPNotFound:
            raise ProjectNotFound(project_id)

    def get_run_url(
        self, run_id: str, workspace: str, project_name: str, short_id: str
    ) -> str:
        base_url = self.get_display_address()
        return f"{base_url}/{workspace}/{project_name}/e/{short_id}"

    @staticmethod
    def _get_all_items(get_portion, step):
        items = []
        previous_items = None
        while previous_items is None or len(previous_items) >= step:
            previous_items = get_portion(limit=step, offset=len(items))
            items += previous_items
        return items



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import os
import unittest
import uuid
from datetime import datetime

from mock import Mock, patch

from neptune.new import ANONYMOUS, Run, get_last_run, get_project, init
from neptune.new.attributes.atoms import String
from neptune.new.envs import API_TOKEN_ENV_NAME, PROJECT_ENV_NAME
from neptune.new.exceptions import (
    MetadataInconsistency,
    NeptuneOfflineModeFetchException,
    NeptuneUninitializedException,
)
from neptune.new.internal.backends.api_model import (
    ApiRun,
    Attribute,
    AttributeType,
    AttributeWithProperties,
    LeaderboardEntry,
    IntAttribute,
)
from neptune.new.internal.backends.neptune_backend_mock import NeptuneBackendMock
from neptune.utils import IS_WINDOWS


@patch("neptune.new.internal.backends.factory.HostedNeptuneBackend", NeptuneBackendMock)
class TestClient(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        os.environ[PROJECT_ENV_NAME] = "organization/project"
        os.environ[API_TOKEN_ENV_NAME] = ANONYMOUS

    def test_incorrect_mode(self):
        with self.assertRaises(ValueError):
            init(mode="srtgj")

    def test_debug_mode(self):
        exp = init(mode="debug")
        exp["some/variable"] = 13
        self.assertEqual(13, exp["some/variable"].fetch())
        self.assertNotIn(str(exp._id), os.listdir(".neptune"))

    def test_offline_mode(self):
        exp = init(mode="offline")
        exp["some/variable"] = 13
        with self.assertRaises(NeptuneOfflineModeFetchException):
            exp["some/variable"].fetch()
        self.assertIn(str(exp._id), os.listdir(".neptune/offline"))
        self.assertIn("data-1.log", os.listdir(".neptune/offline/{}".format(exp._id)))

    def test_sync_mode(self):
        exp = init(mode="sync")
        exp["some/variable"] = 13
        self.assertEqual(13, exp["some/variable"].fetch())
        self.assertNotIn(str(exp._id), os.listdir(".neptune"))

    def test_async_mode(self):
        with init(mode="async", flush_period=0.5) as exp:
            exp["some/variable"] = 13
            with self.assertRaises(MetadataInconsistency):
                exp["some/variable"].fetch()
            exp.wait()
            self.assertEqual(13, exp["some/variable"].fetch())
            self.assertIn(str(exp._id), os.listdir(".neptune/async"))
            execution_dir = os.listdir(".neptune/async/{}".format(exp._id))[0]
            self.assertIn(
                "data-1.log",
                os.listdir(".neptune/async/{}/{}".format(exp._id, execution_dir)),
            )

    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_run",
        new=lambda _, _id: ApiRun(
            uuid.UUID("12345678-1234-5678-1234-567812345678"),
            "SAN-94",
            "workspace",
            "sandbox",
            False,
        ),
    )
    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_attributes",
        new=lambda _, _uuid: [Attribute("some/variable", AttributeType.INT)],
    )
    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_int_attribute",
        new=lambda _, _uuid, _path: IntAttribute(42),
    )
    def test_read_only_mode(self):
        exp = init(mode="read-only", run="SAN-94")

        with self.assertLogs() as caplog:
            exp["some/variable"] = 13
            exp["some/other_variable"] = 11
            self.assertEqual(
                caplog.output,
                [
                    "WARNING:neptune.new.internal.operation_processors.read_only_operation_processor:"
                    "Client in read-only mode, nothing will be saved to server."
                ],
            )

        self.assertEqual(42, exp["some/variable"].fetch())
        self.assertNotIn(str(exp._id), os.listdir(".neptune"))

    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_run",
        new=lambda _, _id: ApiRun(
            uuid.UUID("12345678-1234-5678-1234-567812345678"),
            "SAN-94",
            "workspace",
            "sandbox",
            False,
        ),
    )
    @patch(
        "neptune.new.internal.backends.neptune_backend_mock.NeptuneBackendMock.get_attributes",
        new=lambda _, _uuid: [Attribute("test", AttributeType.STRING)],
    )
    def test_resume(self):
        with init(flush_period=0.5, run="SAN-94") as exp:
            self.assertEqual(exp._id, uuid.UUID("12345678-1234-5678-1234-567812345678"))
            self.assertIsInstance(exp.get_structure()["test"], String)

    @patch("neptune.new.internal.utils.source_code.sys.argv", ["main.py"])
    @patch(
        "neptune.new.internal.init_impl.os.path.isfile", new=lambda file: "." in file
    )
    @patch(
        "neptune.new.internal.utils.glob",
        new=lambda path, recursive=False: [path.replace("*", "file.txt")],
    )
    @patch(
        "neptune.new.internal.utils.os.path.abspath",
        new=lambda path: os.path.normpath("/home/user/main_dir/" + path),
    )
    @patch("neptune.new.internal.utils.os.getcwd", new=lambda: "/home/user/main_dir")
    @unittest.skipIf(IS_WINDOWS, "Linux/Mac test")
    def test_entrypoint(self):
        exp = init(mode="debug")
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main.py")

        exp = init(mode="debug", source_files=[])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main.py")

        exp = init(mode="debug", source_files=["../*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main_dir/main.py")

        exp = init(mode="debug", source_files=["internal/*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "main.py")

        exp = init(mode="debug", source_files=["../other_dir/*"])
        self.assertEqual(exp["source_code/entrypoint"].fetch(), "../main_dir/main.py")

    @patch("neptune.new.internal.utils.source_code.sys.argv", ["main.py"])
    @patch("neptune.new.internal.utils.source_code.is_ipython", new=lambda: True)
    def test_entrypoint_in_interactive_python(self):
        exp = init(mode="debug")
        with self.assertRaises(AttributeError):
            exp["source_code/entrypoint"].get()

        exp = init(mode="debug", source_files=[])
        with self.assertRaises(AttributeError):
            exp["source_code/entrypoint"].get()

        exp = init(mode="debug", source_files=["../*"])
        with self.assertRaises(AttributeError):
            exp["source_code/entrypoint"].get()

        exp = init(mode="debug", source_files=["internal/*"])
        with self.assertRaises(AttributeError):
            exp["source_code/entrypoint"].get()

    @patch("neptune.new.internal.utils.source_code.sys.argv", ["main.py"])
    @patch("neptune.new.internal.utils.source_code.get_common_root", new=lambda _: None)
    @patch(
        "neptune.new.internal.init_impl.os.path.isfile", new=lambda file: "." in file
    )
    @patch(
        "neptune.new.internal.utils.glob",
        new=lambda path, recursive=False: [path.replace("*", "file.txt")],
    )
    @patch(
        "neptune.new.internal.utils.os.path.abspath",
        new=lambda path: os.path.normpath("/home/user/main_dir/" + path),
    )
    def test_entrypoint_without_common_root(self):
        exp = init(mode="debug", source_files=["../*"])
        self.assertEqual(
            exp["source_code/entrypoint"].fetch(), "/home/user/main_dir/main.py"
        )

        exp = init(mode="debug", source_files=["internal/*"])
        self.assertEqual(
            exp["source_code/entrypoint"].fetch(), "/home/user/main_dir/main.py"
        )

    @patch("neptune.new.internal.get_project_impl.HostedNeptuneBackend")
    def test_get_table_as_pandas(self, backend_init_mock):
        # given
        backend_mock = Mock()
        backend_init_mock.return_value = backend_mock

        # and
        attributes = []
        now = datetime.now()
        attributes.append(
            AttributeWithProperties(
                "run/state", AttributeType.RUN_STATE, Mock(value="idle")
            )
        )
        attributes.append(
            AttributeWithProperties("float", AttributeType.FLOAT, Mock(value=12.5))
        )
        attributes.append(
            AttributeWithProperties(
                "string", AttributeType.STRING, Mock(value="some text")
            )
        )
        attributes.append(
            AttributeWithProperties("datetime", AttributeType.DATETIME, Mock(value=now))
        )
        attributes.append(
            AttributeWithProperties(
                "float/series", AttributeType.FLOAT_SERIES, Mock(last=8.7)
            )
        )
        attributes.append(
            AttributeWithProperties(
                "string/series", AttributeType.STRING_SERIES, Mock(last="last text")
            )
        )
        attributes.append(
            AttributeWithProperties(
                "string/set", AttributeType.STRING_SET, Mock(values=["a", "b"])
            )
        )
        attributes.append(
            AttributeWithProperties(
                "git/ref",
                AttributeType.GIT_REF,
                Mock(commit=Mock(commitId="abcdef0123456789")),
            )
        )
        attributes.append(AttributeWithProperties("file", AttributeType.FILE, None))
        attributes.append(
            AttributeWithProperties("file/set", AttributeType.FILE_SET, None)
        )
        attributes.append(
            AttributeWithProperties("image/series", AttributeType.IMAGE_SERIES, None)
        )

        # and
        empty_entry = LeaderboardEntry(uuid.uuid4(), [])
        filled_entry = LeaderboardEntry(uuid.uuid4(), attributes)
        backend_mock.get_leaderboard = Mock(return_value=[empty_entry, filled_entry])

        # when
        df = get_project().fetch_runs_table().to_pandas()

        # then
        self.assertEqual("idle", df["run/state"][1])
        self.assertEqual(12.5, df["float"][1])
        self.assertEqual("some text", df["string"][1])
        self.assertEqual(now, df["datetime"][1])
        self.assertEqual(8.7, df["float/series"][1])
        self.assertEqual("last text", df["string/series"][1])
        self.assertEqual("a,b", df["string/set"][1])
        self.assertEqual("abcdef0123456789", df["git/ref"][1])

        with self.assertRaises(KeyError):
            self.assertTrue(df["file"])
        with self.assertRaises(KeyError):
            self.assertTrue(df["file/set"])
        with self.assertRaises(KeyError):
            self.assertTrue(df["image/series"])

    @patch("neptune.new.internal.get_project_impl.HostedNeptuneBackend")
    def test_get_table_as_runs(self, backend_init_mock):
        # given
        backend_mock = Mock()
        backend_init_mock.return_value = backend_mock

        # and
        exp_id = uuid.uuid4()
        attributes = []
        now = datetime.now()
        attributes.append(
            AttributeWithProperties(
                "run/state", AttributeType.RUN_STATE, Mock(value="idle")
            )
        )
        attributes.append(
            AttributeWithProperties("float", AttributeType.FLOAT, Mock(value=12.5))
        )
        attributes.append(
            AttributeWithProperties(
                "string", AttributeType.STRING, Mock(value="some text")
            )
        )
        attributes.append(
            AttributeWithProperties("datetime", AttributeType.DATETIME, Mock(value=now))
        )
        attributes.append(
            AttributeWithProperties(
                "float/series", AttributeType.FLOAT_SERIES, Mock(last=8.7)
            )
        )
        attributes.append(
            AttributeWithProperties(
                "string/series", AttributeType.STRING_SERIES, Mock(last="last text")
            )
        )
        attributes.append(
            AttributeWithProperties(
                "string/set", AttributeType.STRING_SET, Mock(values=["a", "b"])
            )
        )
        attributes.append(
            AttributeWithProperties(
                "git/ref",
                AttributeType.GIT_REF,
                Mock(commit=Mock(commitId="abcdef0123456789")),
            )
        )
        attributes.append(AttributeWithProperties("file", AttributeType.FILE, None))
        attributes.append(
            AttributeWithProperties("file/set", AttributeType.FILE_SET, None)
        )
        attributes.append(
            AttributeWithProperties("image/series", AttributeType.IMAGE_SERIES, None)
        )

        # and
        backend_mock.get_leaderboard = Mock(
            return_value=[LeaderboardEntry(exp_id, attributes)]
        )

        # when
        exp = get_project().fetch_runs_table().to_runs()[0]

        # then
        self.assertEqual("idle", exp["run/state"].get())
        self.assertEqual("idle", exp["run"]["state"].get())
        self.assertEqual(12.5, exp["float"].get())
        self.assertEqual("some text", exp["string"].get())
        self.assertEqual(now, exp["datetime"].get())
        self.assertEqual(8.7, exp["float/series"].get())
        self.assertEqual("last text", exp["string/series"].get())
        self.assertEqual({"a", "b"}, exp["string/set"].get())
        self.assertEqual("abcdef0123456789", exp["git/ref"].get())

        with self.assertRaises(MetadataInconsistency):
            exp["file"].get()
        with self.assertRaises(MetadataInconsistency):
            exp["file/set"].get()
        with self.assertRaises(MetadataInconsistency):
            exp["image/series"].get()

        exp["file"].download("some_directory")
        backend_mock.download_file.assert_called_with(
            exp_id, ["file"], "some_directory"
        )

        exp["file/set"].download("some_directory")
        backend_mock.download_file_set.assert_called_with(
            exp_id, ["file", "set"], "some_directory"
        )

    def test_last_exp_is_raising_exception_when_non_initialized(self):
        # given uninitialized run
        Run.last_run = None

        # expect: raises NeptuneUninitializedException
        with self.assertRaises(NeptuneUninitializedException):
            get_last_run()

    def test_last_exp_is_the_latest_initialized(self):
        # given two initialized runs
        with init() as exp1, init() as exp2:
            # expect: `neptune.latest_run` to be the latest initialized one
            self.assertIsNot(exp1, get_last_run())
            self.assertIs(exp2, get_last_run())



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import io
import tarfile
from typing import AnyStr, Union, BinaryIO, Any, Generator

from future.builtins import object

from neptune.internal.storage.storage_utils import (
    UploadEntry,
    AttributeUploadConfiguration,
)


class FileChunk(object):
    def __init__(self, data: AnyStr, start, end):
        self.data = data
        self.start = start
        self.end = end

    def get_data(self) -> AnyStr:
        return self.data

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class FileChunkStream(object):
    def __init__(
        self,
        upload_entry: UploadEntry,
        upload_configuration: AttributeUploadConfiguration,
    ):
        self.filename: str = upload_entry.target_path
        self.upload_configuration: AttributeUploadConfiguration = upload_configuration
        self.length: int = upload_entry.length()
        self.fobj: Union[BinaryIO, io.BytesIO] = upload_entry.get_stream()
        self.permissions: str = upload_entry.get_permissions()

    def __eq__(self, fs):
        if isinstance(self, fs.__class__):
            return self.__dict__ == fs.__dict__
        return False

    def generate(self) -> Generator[FileChunk, Any, None]:
        last_offset = 0
        while True:
            chunk = self.fobj.read(self.upload_configuration.chunk_size)
            if chunk:
                if isinstance(chunk, str):
                    chunk = chunk.encode("utf-8")
                new_offset = last_offset + len(chunk)
                yield FileChunk(chunk, last_offset, new_offset)
                last_offset = new_offset
            else:
                if last_offset == 0:
                    yield FileChunk(b"", 0, 0)
                break

    def close(self):
        self.fobj.close()


def compress_to_tar_gz_in_memory(upload_entries) -> bytes:
    f = io.BytesIO(b"")

    with tarfile.TarFile.open(fileobj=f, mode="w|gz", dereference=True) as archive:
        for entry in upload_entries:
            archive.add(
                name=entry.source_path, arcname=entry.target_path, recursive=True
            )

    f.seek(0)
    data = f.read()
    return data
